from collections.abc import Hashable
from enum import Enum, auto, unique
from logging import getLogger

from chatushka.core.matchers.base import MatcherBase
from chatushka.core.models import MatchedToken, Update

logger = getLogger(__name__)


@unique
class ChatUsersMovementsEventsEnum(Enum):
    CAME = auto()


class ChatUsersMovementsMatcher(MatcherBase):
    def _cast_token(
        self,
        token: str | ChatUsersMovementsEventsEnum,
    ) -> ChatUsersMovementsEventsEnum:
        if isinstance(token, str):
            return ChatUsersMovementsEventsEnum[token.upper()]
        return token

    async def _check(
        self,
        token: Hashable,
        update: Update,
    ) -> MatchedToken | None:
        if update.message and update.message.new_chat_members:
            return MatchedToken(token=ChatUsersMovementsEventsEnum.CAME)
        return None



from logging import getLogger
from re import findall

from chatushka.core.matchers.base import MatcherBase
from chatushka.core.models import MatchedToken, RegexMatchKwargs, Update

logger = getLogger(__name__)


class RegexMatcher(MatcherBase):
    suffix = "regex"

    async def _check(
        self,
        token: str,  # type: ignore
        update: Update,
    ) -> MatchedToken | None:
        if not update.message or not update.message.text:
            return None
        if founded := findall(token, update.message.text):
            kwargs = RegexMatchKwargs(matched=tuple(founded))
            return MatchedToken(
                token=token,
                kwargs=kwargs,
            )
        return None



from collections.abc import Callable, Coroutine, Hashable, Iterable
from typing import Any, Protocol

from chatushka.core.models import HANDLER_TYPING, MatchedToken, Message
from chatushka.core.telegram import Telegram


class MatcherProtocol(Protocol):
    handlers: dict[Hashable, list[HANDLER_TYPING]]

    def __call__(
        self,
        *tokens: Hashable,
    ) -> Callable[[Callable[[], None]], None]:
        ...

    def add_handler(
        self,
        tokens: Hashable | Iterable[Hashable],
        handler: Callable[[], None] | Callable[[], Coroutine],  # type: ignore
    ) -> None:
        ...

    async def check_handlers(
        self,
        message: Message,
    ) -> None:
        ...

    async def match(
        self,
        api: Telegram,
        message: Message,
    ) -> MatchedToken:
        ...

    async def call(
        self,
        api: Telegram,
        token: Hashable,
        message: Message | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None:
        ...

    async def init(self) -> None:
        ...



from chatushka import CommandsMatcher, Message, Telegram
from chatushka.bot.settings import get_settings

settings = get_settings()
helpers_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)


@helpers_matcher("id")
async def id_handler(
    api: Telegram,
    message: Message,
) -> None:
    admins = await api.get_chat_administrators(message.chat.id)
    line_tmpl = "{id_type}: <pre>{id_value}</pre>"
    ids = {"user_id": message.user.id}
    for admin in admins:
        if admin.user.id == message.user.id:
            ids = ids | {"chat_id": message.chat.id}
    text = "\n".join([line_tmpl.format(id_type=id_type, id_value=id_value) for id_type, id_value in ids.items()])
    await api.send_message(
        chat_id=message.chat.id,
        text=text,
        reply_to_message_id=message.message_id,
    )


@helpers_matcher("ping", "пинг")
async def ping_handler(
    api: Telegram,
    message: Message,
) -> None:
    answer = "pong" if "ping" in message.text else "понг"
    await api.send_message(
        chat_id=message.chat.id,
        text=answer,
        reply_to_message_id=message.message_id,
    )



from datetime import timedelta
from random import randrange

from chatushka import CommandsMatcher, Message, Telegram
from chatushka.bot.matchers.admin.utils import send_mute_request
from chatushka.bot.settings import get_settings

settings = get_settings()
mute_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)


@mute_matcher("mute", "shutup", "kill")
async def mute_handler(
    api: Telegram,
    message: Message,
    args: list[str],
) -> None:
    restrict_time = timedelta(minutes=randrange(10, 30))
    admins = await api.get_chat_administrators(message.chat.id)

    privileges = None
    for admin in admins:
        if admin.user.id == message.user.id:
            privileges = admin
            break

    if (
        not message.reply_to_message
        or not privileges
        or not privileges.status.CREATOR
        or not privileges.can_restrict_members
    ):
        await send_mute_request(
            api=api,
            message=message,
            initiator=message.user,
            restrict_user=message.user,
            restrict_time=restrict_time,
        )
        return

    try:  # noqa
        restrict_time = timedelta(hours=int(args[0]))
    except (ValueError, IndexError):
        pass

    await send_mute_request(
        api=api,
        message=message,
        initiator=message.user,
        restrict_user=message.reply_to_message.user,
        restrict_time=restrict_time,
    )



from collections.abc import Hashable, Iterable

from chatushka.core.matchers.base import MatcherBase
from chatushka.core.models import MatchedToken, Update


class CommandsMatcher(MatcherBase):
    def __init__(
        self,
        prefixes: str | tuple[str, ...] = ("/",),
        postfixes: str | tuple[str, ...] = (),
        allow_raw: bool = False,
        case_sensitive: bool = False,
        whitelist: tuple[int, ...] | None = None,
    ) -> None:
        super().__init__()
        if isinstance(prefixes, str):
            prefixes = (prefixes,)
        if isinstance(postfixes, str):
            postfixes = (postfixes,)

        variations = [prefix + "{cmd}" for prefix in prefixes if prefix.strip()] + [
            "{cmd}" + postfix for postfix in postfixes if postfix.strip()
        ]
        if allow_raw:
            variations.append("{cmd}")

        self._variations = set(variations)
        self._case_sensitive = case_sensitive
        self._whitelist = whitelist

    def _cast_token(
        self,
        token: Hashable,
    ) -> Hashable | Iterable[Hashable]:
        tokens = []
        for variation in self._variations:
            value = variation.format(cmd=token)
            if not self._case_sensitive:
                value = value.lower()
            tokens.append(value)
        return tokens

    async def _check(
        self,
        token: str,
        update: Update,
    ) -> MatchedToken | None:
        if not update.message or not update.message.text:
            return None
        if self._whitelist and update.message.user.id not in self._whitelist:
            return None
        words = tuple(word for word in update.message.text.split(" ") if word)
        for i, word in enumerate(words):
            if not self._case_sensitive:
                word = word.lower()
            if token == word:
                return MatchedToken(
                    token=token,
                    args=tuple(words[i + 1 :]),
                )
        return None



from collections.abc import Hashable, Iterable
from logging import getLogger

from chatushka.core.matchers.base import MatcherBase
from chatushka.core.models import EventTypes, MatchedToken, Update

logger = getLogger(__name__)


class EventsMatcher(MatcherBase):
    def _cast_token(
        self,
        token: str,
    ) -> str | Iterable[str]:
        if isinstance(token, str):
            return EventTypes[token.upper()]
        return token

    async def _check(
        self,
        token: Hashable,
        update: Update,
    ) -> MatchedToken | None:
        return MatchedToken(token=EventTypes.MESSAGE)



from contextlib import suppress
from logging import getLogger

from chatushka.core.matchers.base import MatcherBase

with suppress(ImportError):
    from aiocron import crontab


logger = getLogger(__name__)


class CronMatcher(MatcherBase):
    async def init(self) -> None:
        for token, handlers in self.handlers.items():
            for handler in handlers:
                crontab(token, func=handler)



from functools import lru_cache
from typing import Any

from yaml import safe_load

from chatushka.bot.settings import BOT_DATA_DIR


@lru_cache
def read_yaml_from_data_dir(
    filename: str,
) -> dict[str, Any]:
    if not filename.endswith("yaml"):
        filename += ".yaml"
    path = BOT_DATA_DIR / filename
    with path.open(encoding="utf8") as fh:
        data = fh.read()
    return safe_load(data)


def read_txt_from_data_dir(
    filename: str,
) -> list[str]:
    if not filename.endswith("txt"):
        filename += ".txt"
    path = BOT_DATA_DIR / filename
    with path.open(encoding="utf8") as fh:
        data = fh.read()
    return data.strip().split("\n")



from chatushka import CommandsMatcher
from chatushka.bot.matchers.admin.mute import mute_matcher
from chatushka.bot.matchers.admin.pin import pin_matcher
from chatushka.bot.settings import get_settings

settings = get_settings()
admin_matcher = CommandsMatcher(
    prefixes=settings.command_prefixes,
    postfixes=settings.command_postfixes,
)
admin_matcher.add_matcher(
    mute_matcher,
    pin_matcher,
)

__all__ = ("admin_matcher",)



from abc import ABC
from asyncio import iscoroutinefunction
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable
from inspect import signature
from typing import Any, NamedTuple

from chatushka import MatcherProtocol, Telegram
from chatushka.core.models import HANDLER_TYPING, MatchedToken, Update


class HelpMessage(NamedTuple):
    tokens: tuple[str, ...]
    message: str | None


class MatcherBase(ABC):  # noqa
    def __init__(
        self,
    ) -> None:
        self.handlers: dict[Hashable, list[HANDLER_TYPING]] = defaultdict(list)
        self.matchers: list[MatcherProtocol] = []
        self._help_messages: list[HelpMessage] = []

    def __call__(
        self,
        *tokens: Hashable,
        help_message: str | None = None,
        include_in_help: bool = True,
    ) -> Callable[[Callable[[], None]], None]:
        def decorator(
            func: HANDLER_TYPING,
        ) -> None:
            self.add_handler(
                tokens=tokens,
                handler=func,
                help_message=help_message,
                include_in_help=include_in_help,
            )

        return decorator

    @property
    def help_messages(self) -> list[HelpMessage]:
        messages = self._help_messages.copy()
        for matcher in self.matchers:
            messages += matcher.help_messages
        return messages

    def add_handler(
        self,
        tokens: Hashable | Iterable[Hashable],
        handler: HANDLER_TYPING,
        help_message: str | None = None,
        include_in_help: bool = True,
    ) -> None:
        if not help_message:
            help_message = f"help message of {self.__class__.__name__}"
        if not isinstance(tokens, list | tuple | set):
            tokens = (tokens,)
        for raw_token in tokens:
            if isinstance(raw_token, str):
                raw_token = raw_token.strip()
            prepared = self._cast_token(raw_token)
            if not isinstance(prepared, list | tuple | set):
                prepared = (prepared,)
            for token in prepared:
                self.handlers[token].append(handler)
        if include_in_help:
            self._help_messages.append(
                HelpMessage(tokens, help_message),
            )

    def add_matcher(
        self,
        *matchers: MatcherProtocol,
    ) -> None:
        self.matchers += matchers

    async def match(
        self,
        api: Telegram,
        update: Update,
        *,
        should_call_matched: bool = False,
    ) -> list[MatchedToken]:
        matched_handlers = []
        for token in self.handlers:
            if matched := await self._check(token, update):
                matched_handlers.append(matched)
                if should_call_matched:
                    await self.call(
                        api=api,
                        token=matched.token,
                        update=update,
                        kwargs=matched.kwargs | {"args": matched.args},
                    )
        for matcher in self.matchers:
            matched_handlers += await matcher.match(api, update, should_call_matched=should_call_matched)
        return matched_handlers

    async def call(
        self,
        api: Telegram,
        token: Hashable,
        update: Update | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> None:
        if not kwargs:
            kwargs = {}
        kwargs = kwargs | {"api": api, "update": update, "token": token}
        handlers = self.handlers.get(token)
        if not handlers:
            return
        for handler in handlers:
            sig = signature(handler)
            sig_kwargs = {param: kwargs.get(param) for param in sig.parameters if param in kwargs}
            if update and update.message is not None and "message" in sig.parameters:
                sig_kwargs["message"] = update.message
            if iscoroutinefunction(handler):
                await handler(**sig_kwargs)  # type: ignore
                continue
            handler(**sig_kwargs)

    # pylint: disable=no-self-use
    def _cast_token(
        self,
        token: Hashable,
    ) -> Any | Iterable[Any]:
        return (token,)

    # pylint: disable=unused-argument
    async def _check(
        self,
        token: Hashable,
        update: Update,
    ) -> MatchedToken | None:
        return None

    async def init(self) -> None:  # noqa
        pass
